#!/usr/bin/env python3

"""

    Copyright (c) 2017 Martin F. Falatic

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

"""

import os
import sys
import stat
import ctypes
import logging
import mmap
import multiprocessing
import dirtreedigest.utils as dtutils
import dirtreedigest.worker as dtworker
import dirtreedigest.digester as dtdigester

# pylint: disable=bad-whitespace
class Walker(object):
    """ Directory walker and supporting functions """

    # Windows file attributes
    FILE_ATTRIBUTE_NONE                = 0x00000
    FILE_ATTRIBUTE_READONLY            = 0x00001
    FILE_ATTRIBUTE_HIDDEN              = 0x00002
    FILE_ATTRIBUTE_SYSTEM              = 0x00004
    FILE_ATTRIBUTE_RESERVED_0008       = 0x00008
    FILE_ATTRIBUTE_DIRECTORY           = 0x00010
    FILE_ATTRIBUTE_ARCHIVE             = 0x00020
    FILE_ATTRIBUTE_DEVICE              = 0x00040
    FILE_ATTRIBUTE_NORMAL              = 0x00080
    FILE_ATTRIBUTE_TEMPORARY           = 0x00100
    FILE_ATTRIBUTE_SPARSE_FILE         = 0x00200
    FILE_ATTRIBUTE_REPARSE_POINT       = 0x00400
    FILE_ATTRIBUTE_COMPRESSED          = 0x00800
    FILE_ATTRIBUTE_OFFLINE             = 0x01000
    FILE_ATTRIBUTE_NOT_CONTENT_INDEXED = 0x02000  # pylint: disable=invalid-name
    FILE_ATTRIBUTE_ENCRYPTED           = 0x04000
    FILE_ATTRIBUTE_INTEGRITY_STREAM    = 0x08000
    FILE_ATTRIBUTE_VIRTUAL             = 0x10000
    FILE_ATTRIBUTE_NO_SCRUB_DATA       = 0x20000
    FILE_ATTRIBUTE_INVALID             = 0x0FFFF
# pylint: enable=bad-whitespace

    def __init__(self):
        self.logger = logging.getLogger('walker')

    def start_workers(self, control_data):
        """ Start long-running worker processes """
        control_data['p_worker_procs'] = []
        control_data['q_work_units'] = []
        control_data['q_results'] = multiprocessing.Queue()
        control_data['q_debug'] = multiprocessing.Queue()
        control_data['buffer_blocks'] = []
        control_data['buffer_sizes'] = []
        control_data['buffer_names'] = []
        control_data['ignored_file_pats'] = dtutils.compile_patterns(
            control_data['ignored_files'],
            control_data['ignore_path_case'],
        )
        control_data['ignored_dir_pats'] = dtutils.compile_patterns(
            control_data['ignored_dirs'],
            control_data['ignore_path_case'],
        )
        for i in range(control_data['max_buffers']):
            buffer_name = '{}{}'.format(control_data['mmap_prefix'], i)
            control_data['buffer_names'].append(buffer_name)
            if control_data['mmap_mode']:
                control_data['buffer_blocks'].append(mmap.mmap(
                    -1,
                    control_data['max_block_size'],
                    buffer_name,
                ))
            else:
                control_data['buffer_blocks'].append([None])
            control_data['buffer_sizes'].append(0)
        for i in range(len(control_data['selected_digests'])):
            control_data['q_work_units'].append(multiprocessing.JoinableQueue())
            control_data['p_worker_procs'].append(multiprocessing.Process(
                target=dtworker.worker_process,
                args=(control_data['q_work_units'][i],
                      control_data['q_results'],
                      control_data['q_debug'],
                      control_data['mmap_mode'],
                     )
            ))
            #control_data['p_worker_procs'][i].daemon = True
            control_data['p_worker_procs'][i].start()
        # Until workers are ended, raising exceptions can hang the parent process

    def end_workers(self, control_data):
        """ Signal processes that they're all done """
        for q_work_unit in control_data['q_work_units']:
            q_work_unit.put((
                dtworker.WorkerSignal.QUIT,
                None,
                None,
            ))
        while not control_data['q_results'].empty():
            retval = control_data['q_results'].get(True)
            self.logger.debug('Draining queue: %s', retval)
        while any(control_data['p_worker_procs']):
            for i in range(len(control_data['selected_digests'])):
                if (control_data['p_worker_procs'][i] is not None) and (
                        not control_data['p_worker_procs'][i].is_alive()):
                    self.logger.debug(
                        'join %d (state = %s) at %f',
                        i,
                        control_data['p_worker_procs'][i].is_alive(),
                        dtutils.curr_time_secs()
                    )
                    control_data['p_worker_procs'][i].join()
                    control_data['p_worker_procs'][i] = None
        dtutils.flush_debug_queue(control_data['q_debug'], logging.getLogger('worker'))

    def get_win_filemode(self, elem):
        """ Windows: get system-specific file stats """
        return ctypes.windll.kernel32.GetFileAttributesW(elem)

    def is_win_symlink(self, elem):
        """ Windows: system-specific symlink check """
        return os.path.isdir(elem) and (
            self.get_win_filemode(elem) & self.FILE_ATTRIBUTE_REPARSE_POINT)

    def process_tree(self, control_data):
        """ Process the given directory tree """
        results = []
        self._walk_tree(
            control_data=control_data,
            root_dir=control_data['root_dir'],
            callback=self.visit_element,
            results=results)
        return results

    def _walk_tree(self, control_data, root_dir, callback, results):
        """ Re-entrant directory tree walker """
        try:
            dir_list = os.listdir(root_dir)
        except FileNotFoundError:
            self.logger.warning('FileNotFoundError %s', root_dir)
            control_data['counts']['errors'] += 1
            return
        except NotADirectoryError:
            self.logger.warning('NotADirectoryError %s', root_dir)
            control_data['counts']['errors'] += 1
            return
        except PermissionError:
            self.logger.warning('PermissionError %s', root_dir)
            control_data['counts']['errors'] += 1
            return
        for elem in sorted(dir_list):
            pathname = dtutils.unixify_path(os.path.join(root_dir, elem))
            print("Processing {}".format(pathname))
            stats = os.lstat(pathname)
            if stat.S_ISDIR(stats.st_mode):
                if dtutils.elem_is_matched(
                        root_dir,
                        pathname,
                        control_data['ignored_dir_pats']):
                    self.logger.info('D IGNORED %s', pathname)
                    control_data['counts']['ignored'] += 1
                    continue
                results.append(self.visit_element(control_data, pathname, stats))
                self._walk_tree(
                    control_data=control_data,
                    root_dir=pathname,
                    callback=callback,
                    results=results)
            elif stat.S_ISREG(stats.st_mode):
                if dtutils.elem_is_matched(
                        root_dir,
                        pathname,
                        control_data['ignored_file_pats']):
                    self.logger.info('F IGNORED %s', pathname)
                    control_data['counts']['ignored'] += 1
                    continue
                results.append(self.visit_element(control_data, pathname, stats))
            else:
                results.append(self.visit_element(control_data, pathname, stats))
            dtutils.flush_debug_queue(control_data['q_debug'], logging.getLogger('worker'))

    def visit_element(self, control_data, element, stats):
        """ Stat / digest a specific element found during the directory walk """
        root_dir = control_data['root_dir']
        elem_data = {}
        elem_data['name'] = dtutils.get_relative_path(root_dir, dtutils.unixify_path(element))
        elem_data['mode'] = stats.st_mode
        if sys.platform == 'win32':
            elem_data['mode_w'] = self.get_win_filemode(element)
        else:
            elem_data['mode_w'] = self.FILE_ATTRIBUTE_NONE
        elem_data['size'] = stats[stat.ST_SIZE]
        elem_data['atime'] = stats[stat.ST_ATIME]
        elem_data['mtime'] = stats[stat.ST_MTIME]
        elem_data['ctime'] = stats[stat.ST_CTIME]
        elem_data['digests'] = None
        elem_data['type'] = '?'
        alt_digest_len = 1
        if control_data['altfile_digest']:
            alt_digest_len = dtdigester.DIGEST_FUNCTIONS[control_data['altfile_digest']]['len']
        if sys.platform == 'win32' and self.is_win_symlink(element):
            elem_data['type'] = 'J'
            alt_digest = '?' * alt_digest_len
            sorted_digests = dtdigester.fill_digest_str(control_data, 'x')
        elif stat.S_ISDIR(stats.st_mode):
            elem_data['type'] = 'D'
            elem_data['size'] = 0
            alt_digest = '-' * alt_digest_len
            sorted_digests = dtdigester.fill_digest_str(control_data, '-')
            control_data['counts']['dirs'] += 1
        elif stat.S_ISREG(stats.st_mode):
            elem_data['type'] = 'F'
            elem_data['digests'] = dtdigester.digest_file(control_data, element)
            if elem_data['digests']:
                control_data['counts']['files'] += 1
                sorted_digests = '{' + ', '.join('{}: {}'.format(
                    i, elem_data['digests'][i]) for i in sorted(
                        elem_data['digests'])) + '}'
                #self.single_process_digest_test(control_data, element)
                if control_data['altfile_digest']:
                    alt_digest = elem_data['digests'][control_data['altfile_digest']]
            else:
                self.logger.warning('F Problems processing %s', element)
                control_data['counts']['errors'] += 1
                sorted_digests = dtdigester.fill_digest_str(control_data, '!')
                alt_digest = '!' * alt_digest_len
        else:
            elem_data['type'] = '?'
            sorted_digests = dtdigester.fill_digest_str(control_data, '?')
        file_details = '{};{};{:08x};{:08x};{:08x};{:04x};{:04x};{:010x};{}'.format(
            elem_data['type'],
            sorted_digests,
            elem_data['atime'], elem_data['mtime'], elem_data['ctime'],
            elem_data['mode'], elem_data['mode_w'],
            elem_data['size'],
            elem_data['name'])
        alt_digest = '-' * alt_digest_len
        if elem_data['type'] == 'D':
            alt_digest = '-' * alt_digest_len
        elif not elem_data['digests']:
            alt_digest = '?' * alt_digest_len
        elif control_data['altfile_digest']:
            alt_digest = elem_data['digests'][control_data['altfile_digest']]
        alt_details = '{};{:08x};{:08x};{:08x};{:04x};{:010x};{}'.format(
            alt_digest,
            elem_data['atime'], elem_data['mtime'], elem_data['ctime'],
            elem_data['mode_w'],
            elem_data['size'],
            elem_data['name'])
        self.logger.debug('%s', file_details)
        dtutils.outfile_write(control_data['outfile_name'], 'a', [
            '{}'.format(file_details),])
        if control_data['altfile_digest']:
            dtutils.outfile_write(control_data['altfile_name'], 'a', [
                '{}'.format(alt_details),])
        return elem_data

    def single_process_digest_test(self, control_data, element):
        """ Single-process digest for testing """
        digest_name = 'sha1'
        with dtutils.open_with_error_checking(element, 'rb') as (fileh, err):
            if err:
                return None
            mdf = dtdigester.DIGEST_FUNCTIONS[digest_name]['entry']()
            bytes_read = 0
            while True:
                byte_block = fileh.read(control_data['max_block_size'])
                bytes_read += len(byte_block)
                if not byte_block:
                    break
                mdf.update(byte_block)
                self.logger.debug(
                    'single_process_digest_test() processing -- digest=%s l=%d d=%s',
                    digest_name, bytes_read, mdf.hexdigest())
            self.logger.debug(
                'single_process_digest_test() FINAL -- digest=%s l=%d d=%s',
                digest_name, bytes_read, mdf.hexdigest())



#!/usr/bin/env python3

"""

    Copyright (c) 2017 Martin F. Falatic

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

"""

import os
import sys
import hashlib
import zlib
import logging
import dirtreedigest.utils as dtutils
import dirtreedigest.worker as dtworker

class CsumNoop(object):
    """ No-op digester """
    name = 'noop'
    def __init__(self):
        self.checksum = 0
    def update(self, msg):
        """ Update checksum """
        pass
    def hexdigest(self):
        """ Return digest string in hexadecimal format """
        return '{0:0{1}x}'.format(self.checksum, 8)

class CsumNoop1(CsumNoop):
    """ No-op digester """
    name = 'noop1'

class CsumNoop2(CsumNoop):
    """ No-op digester """
    name = 'noop2'

class CsumNoop3(CsumNoop):
    """ No-op digester """
    name = 'noop3'

class CsumNoop4(CsumNoop):
    """ No-op digester """
    name = 'noop4'

class CsumNoop5(CsumNoop):
    """ No-op digester """
    name = 'noop5'

class CsumNoop6(CsumNoop):
    """ No-op digester """
    name = 'noop6'

class CsumNoop7(CsumNoop):
    """ No-op digester """
    name = 'noop7'

class CsumNoop8(CsumNoop):
    """ No-op digester """
    name = 'noop8'

class CsumAdler32(object):
    """ Adler32 digester """
    name = 'adler32'
    def __init__(self):
        self.checksum = 1
    def update(self, msg):
        """ Update checksum """
        self.checksum = zlib.adler32(msg, self.checksum)
    def hexdigest(self):
        """ Return digest string in hexadecimal format """
        return '{0:0{1}x}'.format(self.checksum, 8)

class CsumCrc32(object):
    """ CRC32 digester """
    name = 'crc32'
    def __init__(self):
        self.checksum = 0
    def update(self, msg):
        """ Update checksum """
        self.checksum = zlib.crc32(msg, self.checksum)
    def hexdigest(self):
        """ Return digest string in hexadecimal format """
        return '{0:0{1}x}'.format(self.checksum, 8)

# pylint: disable=bad-whitespace, no-member
DIGEST_FUNCTIONS = {
    'noop':       {'name': 'noop',      'len':   8, 'entry': CsumNoop},
    'noop1':      {'name': 'noop1',     'len':   8, 'entry': CsumNoop1},
    'noop2':      {'name': 'noop2',     'len':   8, 'entry': CsumNoop2},
    'noop3':      {'name': 'noop3',     'len':   8, 'entry': CsumNoop3},
    'noop4':      {'name': 'noop4',     'len':   8, 'entry': CsumNoop4},
    'noop5':      {'name': 'noop5',     'len':   8, 'entry': CsumNoop5},
    'noop6':      {'name': 'noop6',     'len':   8, 'entry': CsumNoop6},
    'noop7':      {'name': 'noop7',     'len':   8, 'entry': CsumNoop7},
    'noop8':      {'name': 'noop8',     'len':   8, 'entry': CsumNoop8},
    'crc32':      {'name': 'crc32',     'len':   8, 'entry': CsumCrc32},
    'adler32':    {'name': 'adler32',   'len':   8, 'entry': CsumAdler32},
    'md5':        {'name': 'md5',       'len':  32, 'entry': hashlib.md5},
    'sha1':       {'name': 'sha1',      'len':  40, 'entry': hashlib.sha1},
    'sha224':     {'name': 'sha224',    'len':  56, 'entry': hashlib.sha224},
    'sha256':     {'name': 'sha256',    'len':  64, 'entry': hashlib.sha256},
    'sha384':     {'name': 'sha384',    'len':  96, 'entry': hashlib.sha384},
    'sha512':     {'name': 'sha512',    'len': 128, 'entry': hashlib.sha512},
}
if sys.version_info >= (3,6):
    DIGEST_FUNCTIONS36 = {
        'blake2b':    {'name': 'blake2b',   'len': 128, 'entry': hashlib.blake2b},
        'blake2s':    {'name': 'blake2s',   'len':  64, 'entry': hashlib.blake2s},
        'sha3_224':   {'name': 'sha3_224',  'len':  56, 'entry': hashlib.sha3_224},
        'sha3_256':   {'name': 'sha3_256',  'len':  64, 'entry': hashlib.sha3_256},
        'sha3_384':   {'name': 'sha3_384',  'len':  96, 'entry': hashlib.sha3_384},
        'sha3_512':   {'name': 'sha3_512',  'len': 128, 'entry': hashlib.sha3_512},
        #'shake_128': {'name': 'shake_128', 'len':  32, 'entry': hashlib.shake_128},
        #'shake_256': {'name': 'shake_256', 'len':  64, 'entry': hashlib.shake_256},
    }
    DIGEST_FUNCTIONS.update(DIGEST_FUNCTIONS36)
# pylint: enable=bad-whitespace, no-member

''' From least to most secure, excluding noops '''
DIGEST_PRIORITY = [
    'crc32',
    'adler32',
    'md5',
    'sha1',
    'sha224',
    'sha3_224',
    'sha256',
    'blake2s',
    'sha3_256',
    'sha384',
    'sha3_384',
    'sha512',
    'blake2b',
    'sha3_512',
]

def validate_digests(control_data):
    """ returns list of available digests """
    logger = logging.getLogger('digester')
    if not control_data['selected_digests']:
        return None
    digests_available = set(DIGEST_FUNCTIONS.keys()) & set(control_data['selected_digests'])
    digests_not_found = set(control_data['selected_digests']) - digests_available
    if digests_not_found:
        logger.warning('Warning: invalid digest(s): %s', digests_not_found)
    digest_list = sorted(list(digests_available))
    if len(digest_list) > control_data['max_concurrent_jobs']:
        logger.error(
            'Error: Number of digests (%d) may not exceed max jobs (%d)',
            len(digest_list), control_data['max_concurrent_jobs'])
        return None
    logger.debug('Digests: %s', digest_list)
    return digest_list

def fill_digest_str(control_data, fillchar='-'):
    """ Create a padded dummy digest value """
    return '{' + ', '.join('{}: {}'.format(
        i, fillchar*DIGEST_FUNCTIONS[i]['len']) for i in sorted(
            control_data['selected_digests'])) + '}'

def digest_file(control_data, element):
    """ Digest a given element """
    logger = logging.getLogger('digester')
    start_time = dtutils.curr_time_secs()
    logger.debug('process_file(%s)', element)
    curr_buffer_index = next_buffer_index = 0
    total_jobs = len(control_data['selected_digests'])
    buffers_in_use = 0
    active_jobs = 0
    bytes_read = 0
    found_eof = False
    buffer_full = False
    hash_stats = {}
    file_size = os.path.getsize(element)
    with dtutils.open_with_error_checking(element, 'rb') as (fileh, err):
        if err:
            return None
        for i, q_work_unit in enumerate(control_data['q_work_units']):
            digest_name = control_data['selected_digests'][i]
            digest_func = DIGEST_FUNCTIONS[digest_name]['entry']
            q_work_unit.put((
                dtworker.WorkerSignal.INIT,
                digest_func,
                None,
            ))
        # Process one block buffer completely before the next
        block_read = fileh.read(min(control_data['max_block_size'], file_size-bytes_read))
        if control_data['mmap_mode']:
            control_data['buffer_blocks'][curr_buffer_index].seek(0)
            control_data['buffer_blocks'][curr_buffer_index].write(block_read)
        else:
            control_data['buffer_blocks'][curr_buffer_index] = block_read
        block_len = len(block_read)
        control_data['buffer_sizes'][curr_buffer_index] = block_len
        bytes_read += block_len
        buffers_in_use = 1
        while buffers_in_use:
            active_jobs = 0
            completed_jobs = 0
            next_job = 0
            while completed_jobs < total_jobs: # per buffered block
                while next_job < total_jobs:
                    logger.debug(
                        'QUEUEING JOB %d with buffer #%d',
                        next_job, curr_buffer_index)
                    q_work_unit = control_data['q_work_units'][next_job]
                    if control_data['mmap_mode']:
                        q_work_unit.put((
                            dtworker.WorkerSignal.PROCESS,
                            control_data['buffer_names'][curr_buffer_index],
                            control_data['buffer_sizes'][curr_buffer_index],
                        ))
                    else:
                        # Note: This is noticably more costly versus shared memory mapping
                        q_work_unit.put((
                            dtworker.WorkerSignal.PROCESS,
                            control_data['buffer_blocks'][curr_buffer_index],
                            control_data['buffer_sizes'][curr_buffer_index],
                        ))
                    active_jobs += 1
                    next_job += 1
                    if active_jobs > total_jobs:
                        logger.warning(
                            'active_jobs %d > digests %d',
                            active_jobs, total_jobs)
                    if active_jobs == control_data['max_concurrent_jobs']:
                        logger.debug(
                            'active_jobs %d == max_concurrent %d',
                            active_jobs, control_data['max_concurrent_jobs'])
                        break
                logger.debug('ACTIVE JOBS A %d', active_jobs)
                while True:
                    # Prefetch while we wait for active jobs to start completing
                    # We always need at least one buffered item if not eof
                    if not found_eof:
                        new_buffer_index = (next_buffer_index + 1) % control_data['max_buffers']
                        if new_buffer_index == curr_buffer_index:
                            if not buffer_full:
                                logger.debug(
                                    'prefetch buffer is currently full - bufs_in_use %d',
                                    buffers_in_use)
                                buffer_full = True
                        else:
                            next_buffer_index = new_buffer_index
                            buffer_full = False
                            block_read = fileh.read(
                                min(control_data['max_block_size'], file_size-bytes_read))
                            if not block_read:
                                found_eof = True
                                logger.debug('eof found')
                                continue
                            block_len = len(block_read)
                            bytes_read += block_len
                            if control_data['mmap_mode']:
                                control_data['buffer_blocks'][next_buffer_index].seek(0)
                                control_data['buffer_blocks'][next_buffer_index].write(block_read)
                            else:
                                control_data['buffer_blocks'][next_buffer_index] = block_read
                            control_data['buffer_sizes'][next_buffer_index] = block_len
                            buffers_in_use += 1
                            logger.debug(
                                'prefetched %d bytes into buffer #%d bufs_in_use %d (%s)',
                                len(block_read), next_buffer_index,
                                buffers_in_use, element)
                    if not control_data['q_results'].empty():
                        break
                dtutils.flush_debug_queue(control_data['q_debug'], logging.getLogger('worker'))
                while not control_data['q_results'].empty():
                    retval = control_data['q_results'].get(True)
                    logger.debug('Returned: %s', retval)
                    completed_jobs += 1
                    active_jobs -= 1
                logger.debug('ACTIVE JOBS B %d', active_jobs)
            logger.debug(
                'Finished buffer #%d cnt was %d of %d bytes',
                curr_buffer_index,
                buffers_in_use,
                control_data['buffer_sizes'][curr_buffer_index])
            curr_buffer_index = (curr_buffer_index + 1) % control_data['max_buffers']
            buffers_in_use -= 1
        for q_work_unit in control_data['q_work_units']:
            q_work_unit.put((
                dtworker.WorkerSignal.RESULT,
                None,
                None,
            ))
        dtutils.flush_debug_queue(control_data['q_debug'], logging.getLogger('worker'))
        rolloff = total_jobs
        while rolloff > 0:
            retval = control_data['q_results'].get(True)
            logger.debug('RETVAL: %s', retval)
            if not isinstance(retval, tuple):
                logger.warning('retval is not finalized')
                continue
            hash_stats[retval[0]] = retval[1]
            rolloff -= 1
        dtutils.flush_debug_queue(control_data['q_debug'], logging.getLogger('worker'))
    end_time = dtutils.curr_time_secs()
    run_time = end_time-start_time
    logger.debug('MAINLINE finished at %f', end_time)
    logger.debug(
        'run_time= %.3fs rate= %.2f MB/s bytes= %d %s',
        run_time,
        (bytes_read/(1024*1024))/run_time,
        bytes_read,
        element)
    control_data['counts']['bytes_read'] += bytes_read
    return hash_stats



#!/usr/bin/env python3

"""

    Copyright (c) 2017 Martin F. Falatic

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

"""

import re
import logging
from collections import defaultdict
import dirtreedigest.utils as dtutils
import dirtreedigest.digester as dtdigester
from enum import Enum

class DiffType(Enum):
    M_UNDEF = 0  # Undefined (error case)
    M_NONE  = 1  # Diff file Name, Diff Data ("no match" anywhere)
    M_SFSD  = 2  # Same Full name, Same Data (Opposite only)
    M_SFDD  = 3  # Same Full name, Diff Data (Opposite only)
    M_SNSD  = 4  # Same file Name, Same Data (Opposite and Same sides)
    M_DNSD  = 5  # Diff file Name, Same Data (Opposite and Same sides)

class Comparator(object):
    """ Digest blob comparator and supporting functions """
    elements_l = []
    elements_r = []
    files_by_name_l = {}
    files_by_name_r = {}
    rootpath_l = ''
    rootpath_r = ''
    best_digest = None
    control_data = None

    def __init__(self, control_data):
        self.logger = logging.getLogger('comparator')
        self.control_data = control_data

    def read_dtd_report(self, filename):
        element_pat = re.compile(
            r"^(.+?);{(.+?)};(.+?);(.+?);(.+?);(.+?);(.+?);(.+?);(.*)$")
        legacy_pat = re.compile(
            r"^(.+?);(.+?);(.+?);(.+?);(.+?);(.+?);(.*)$")
        rootpath_pat = re.compile(
            r"^#\s+Base path:\s+(.*)$")
        elements = []
        rootpath = ''
        with open(filename, 'r', encoding='utf-8') as fileh:
            for line in fileh:
                line = line.rstrip('\n')
                mval = element_pat.match(line)
                if mval:
                    digests = {}
                    for digestpair in mval[2].split(','):
                        (digest, val) = digestpair.strip().split(':')
                        digest = digest.strip()
                        val = val.strip()
                        digests[digest] = val
                    file_type = mval[1]
                    file_name = mval[9]
                    if file_type not in ['D', 'F']:
                        self.logger.warning(
                            "Ignoring file type '%s' for %s",
                            file_type,
                            file_name,
                        )
                    elem = {
                        'type': file_type,
                        'digests': digests,
                        'time_access': mval[3],
                        'time_modified': mval[4],
                        'time_created': mval[5],
                        'attr_std': mval[6],
                        'attr_win': mval[7],
                        'size': mval[8],
                        'fullname': file_name,
                    }
                    elements.append(elem)
                else:
                    mval = rootpath_pat.match(line)
                    if mval:
                        rootpath = mval[1]
                    else:
                        mval = legacy_pat.match(line)
                        if mval:
                            file_name = mval[7]
                            digests = {}
                            digests['md5'] = mval[1]
                            file_type = 'F'
                            if digests['md5'].startswith('?'):
                                file_type = '?'
                            elif digests['md5'].startswith('-'):
                                file_type = 'D'
                            if file_type not in ['D', 'F']:
                                self.logger.warning(
                                    "Ignoring file type '%s' for %s",
                                    file_type,
                                    file_name,
                                )
                            elem = {
                                'type': file_type,
                                'digests': digests,
                                'time_access': mval[2],
                                'time_modified': mval[3],
                                'time_created': mval[4],
                                'attr_std': '0000',
                                'attr_win': mval[5],
                                'size': mval[6],
                                'fullname': mval[7],
                            }
                            elements.append(elem)
        self.filesplit(elements)
        return (rootpath, elements)

    def choose_best_digest_for_compare(self, elems1, elems2):
        best = -1
        if not len(elems1) > 0 and len(elems2) > 0:
            self.logger.error('Cannot choose a digest - not found')
            return None
        if not 'digests' in elems1[0] and 'digests' in elems2[0]:
            self.logger.error('Cannot choose a digest - not found')
            return None
        s1 = set(elems1[0]['digests'])
        s2 = set(elems2[0]['digests'])
        for digest in (s1 & s2):
            if digest in dtdigester.DIGEST_PRIORITY:
                best = max(best, dtdigester.DIGEST_PRIORITY.index(digest))
        if best < 0:
            self.logger.error('Cannot choose a digest - none in common')
            return None
        best_name = dtdigester.DIGEST_PRIORITY[best]
        return best_name

    def filesplit(self, elements):
        for elem in elements:
            if '/' in elem['fullname']:
                (elem['pathpart'], elem['filepart']) = elem['fullname'].rsplit('/', 1)
            else:
                (elem['pathpart'], elem['filepart']) = ('', elem['fullname'])
            #print(path, name)

    def compare(self, file_l, file_r):
        """ Main entry: compare two dirtreedigest reports """
        (self.rootpath_l, self.elements_l) = self.read_dtd_report(file_l)
        (self.rootpath_r, self.elements_r) = self.read_dtd_report(file_r)
        self.best_digest = self.choose_best_digest_for_compare(
            self.elements_l, self.elements_r)
        if self.best_digest is None:
            return None
        (self.files_by_name_l, self.files_by_digest_l) = self.slice_data(
            self.elements_l)
        (self.files_by_name_r, self.files_by_digest_r) = self.slice_data(
            self.elements_r)

        name_set_l = set(self.files_by_name_l)
        name_set_r = set(self.files_by_name_r)
        name_diff_l = name_set_l - name_set_r
        name_diff_r = name_set_r - name_set_l
        name_same = name_set_r & name_set_l

        (elems_changed) = self.compare_by_full_names(name_same)
        (elems_copied, elems_added) = self.check_rhs(name_diff_r)
        (elems_moved, elems_deleted) = self.check_lhs(name_diff_l)

        for elem in sorted(elems_changed, key=lambda k: k['fullname']):
            self.logger.info("CHANGED {}".format(elem['fullname']))
        for elem in sorted(elems_copied, key=lambda k: k['fullname']):
            self.logger.info("COPIED  {} == {}".format(elem['fullname'], '---'))
        for elem in sorted(elems_added, key=lambda k: k['fullname']):
            self.logger.info("ADDED   {}".format(elem['fullname']))
        for elem in sorted(elems_moved, key=lambda k: k['fullname']):
            self.logger.info("MOVED   {} == {}".format(elem['fullname'], '---'))
        for elem in sorted(elems_deleted, key=lambda k: k['fullname']):
            self.logger.info("DELETED {}".format(elem['fullname']))

        self.logger.info("Root L: %s", self.rootpath_l)
        self.logger.info("Root R: %s", self.rootpath_r)
        self.logger.info("BestDG: %s", self.best_digest)
        self.logger.info("ElemsL: %d", len(self.elements_l))
        self.logger.info("ElemsR: %d", len(self.elements_r))
        self.logger.info("FilesL: %d", len(self.files_by_name_l))
        self.logger.info("FilesR: %d", len(self.files_by_name_r))
        self.logger.info("  Both: %d", len(name_same))
        self.logger.info("Only L: %d", len(name_diff_l))
        self.logger.info("Only R: %d", len(name_diff_r))

    def slice_data(self, elements):
        files_by_name = {
            elem['fullname']: elem for elem in elements if elem['type'] == 'F'}
        files_by_digest = defaultdict(list)
        for elem in elements:
            if elem['type'] == 'F':
                cmp_digest = elem['digests'][self.best_digest]
                files_by_digest[cmp_digest].append(elem)
        return (files_by_name, files_by_digest)

    def compare_by_full_names(self, name_same):
        elems_changed = []
        for name in name_same:
            digest_l = self.files_by_name_l[name]['digests'][self.best_digest]
            digest_r = self.files_by_name_r[name]['digests'][self.best_digest]
            self.files_by_name_l[name]['match'] = [self.files_by_digest_r[digest_l]]
            self.files_by_name_r[name]['match'] = [self.files_by_digest_l[digest_r]]
            if digest_l == digest_r:
                self.files_by_name_l[name]['status'] = 'same'
                self.files_by_name_r[name]['status'] = 'same'
            else:
                self.files_by_name_l[name]['status'] = 'changed'
                self.files_by_name_r[name]['status'] = 'changed'
                self.logger.info("Digest changed: %s", name)
                elems_changed.append(self.files_by_name_r[name])
        return (elems_changed)

    def check_rhs(self, name_diff_r):
        elems_copied = []
        elems_added = []
        for name in name_diff_r:
            digest_r = self.files_by_name_r[name]['digests'][self.best_digest]
            #print("checking", name)
            if digest_r in self.files_by_digest_l:
                #print(name, digest_r)
                matched_elems = [elem['fullname'] for elem in self.files_by_digest_l[digest_r]]
                matched_names = ','.join(matched_elems)
                #print("> COPIED  {} == {}".format(name, matched_names))
                self.files_by_name_r[name]['status'] = 'copied'
                for elem in self.files_by_digest_l[digest_r]:
                    if elem['filepart'] == self.files_by_name_r[name]['filepart']:
                        print("Found likely source", self.files_by_name_r[name]['fullname'])
                        break
                self.files_by_name_r[name]['match'] = self.files_by_digest_l[digest_r]
                elems_copied.append(self.files_by_name_r[name])
            else:
                #print("> ADDED   {}".format(name))
                self.files_by_name_r[name]['status'] = 'added'
                elems_added.append(self.files_by_name_r[name])
        return (elems_copied, elems_added)

    def check_lhs(self, name_diff_l):
        elems_moved = []
        elems_deleted = []
        for name in name_diff_l:
            digest_l = self.files_by_name_l[name]['digests'][self.best_digest]
            #print("checking", name)
            if digest_l in self.files_by_digest_r:
                #print(name, digest_l)
                matched_elems = [elem['fullname'] for elem in self.files_by_digest_r[digest_l]]
                matched_names = ','.join(matched_elems)
                #print("< MOVED   {} == {}".format(name, matched_names))
                self.files_by_name_l[name]['status'] = 'moved'
                self.files_by_name_l[name]['match'] = self.files_by_digest_r[digest_l]
                elems_moved.append(self.files_by_name_l[name])
            else:
                #print("< DELETED {}".format(name))
                self.files_by_name_l[name]['status'] = 'deleted'
                elems_deleted.append(self.files_by_name_l[name])
        return (elems_moved, elems_deleted)



#!/usr/bin/env python3

"""

    Copyright (c) 2017 Martin F. Falatic

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

"""

import logging
import re
import os
from contextlib import contextmanager
import time
from datetime import datetime

@contextmanager
def open_with_error_checking(filename, mode='r'):
    """ Open a file for reading with error checking """
    try:
        fileh = open(filename, mode)
    except IOError as err:
        yield None, err
    else:
        try:
            yield fileh, None
        finally:
            fileh.close()

def datetime_as_str(dtm=None):
    if not dtm:
        dtm = datetime.now()
    return "{:%Y%m%d_%H%M%S}".format(dtm)

def unixify_path(path):
    """ Convert path to Unix format """
    return path.replace('\\', '/')

def compile_patterns(patterns, ignorecase=False):
    """ Compile exclusion patterns to regular expressions """
    re_pats = []
    flags = 0
    if ignorecase:
        flags |= re.IGNORECASE
    for pattern in patterns:
        pattern = '^'+re.escape(pattern)+'$'
        re_pats.append(re.compile(pattern, flags=flags))
    return re_pats

def elem_is_matched(root, elem, patterns):
    """ Check if element matches any of the exclusion patterns """
    elem_mod = unixify_path(os.path.normpath(elem))
    rel_elem = get_relative_path(root, elem_mod)
    for re_pat in patterns:
        if re.match(re_pat, rel_elem):
            return True
    return False

def split_net_drive(elem):
    """ For network shares, split network and path parts """
    mval = re.match(r"^(//[^/]+)(.*)$", elem)
    if mval:
        return (mval.group(1), mval.group(2))
    return ('', elem)

def split_win_drive(elem):
    """ For Windows drives, split drive spec and path parts """
    mval = re.match(r"^([a-zA-Z]:)(.*)$", elem)
    if mval:
        return (mval.group(1), mval.group(2))
    return ('', elem)

def get_relative_path(root, elem):
    """ Get the element path relative to a given root path """
    matcher = r'^'+re.escape(unixify_path(root))+r'(.*)$'
    retval = elem
    mval = re.match(matcher, elem)
    if mval:
        retval = mval.group(1)
    if retval != '/':
        retval = retval.strip('/')
    return retval

def compare_paths(path1, path2, ignorecase=False):
    """ Compare two paths """
    path1_mod = unixify_path(os.path.normpath(path1))
    path2_mod = unixify_path(os.path.normpath(path2))
    if ignorecase:
        path1_mod = path1_mod.lower()
        path2_mod = path2_mod.lower()
    return path1_mod == path2_mod

def unix_time_ms(dtm=None):
    """ Get Unix time in msec """
    if dtm is None:
        dtm = datetime.now()
    epoch = datetime.utcfromtimestamp(0)
    return int((dtm - epoch).total_seconds() * 1000.0)

def curr_time_secs():
    """ Get current high-resolution time """
    return time.perf_counter()

def flush_debug_queue(q_debug, logger):
    """ Flush the debug message queue """
    while not q_debug.empty():
        retval = q_debug.get(True)
        log_level = retval[0]
        log_message = retval[1]
        logger.log(log_level, log_message)

def start_logging(filename, log_level, con_level):
    """ Initialize file and console logs """
    logfile_handler = logging.FileHandler(filename, 'w', 'utf-8')
    log_fmt = logging.Formatter('%(levelname)s:%(name)s:%(message)s')
    logfile_handler.setFormatter(log_fmt)
    logfile_handler.setLevel(log_level)
    console_handler = logging.StreamHandler()
    con_fmt = logging.Formatter('%(levelname)s:%(name)s:%(message)s')
    console_handler.setFormatter(con_fmt)
    console_handler.setLevel(con_level)
    logging.basicConfig(
        level=logging.NOTSET,  # This must be set to something
        handlers=[logfile_handler, console_handler])

def outfile_write(fname, fmode, lines):
    """ Write a block of data to the output file """
    with open(fname, fmode, encoding='utf-8') as fileh:
        for line in lines:
            fileh.write('{}\n'.format(line))



#!/usr/bin/env python3

"""

    Copyright (c) 2017 Martin F. Falatic

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

-----------------------------------------------------
    Don't print() directly from anything in here.
-----------------------------------------------------

"""

import os
import mmap
import logging  # For constants - do not log directly from here
from enum import Enum

class WorkerSignal(Enum):
    """ Enums to communicate with workers """
    INIT = 0
    PROCESS = 1
    RESULT = 2
    QUIT = 3

def worker_process(q_work_unit, q_results, q_debug, mmap_mode=False):
    """ This is run as a subprocess, potentially with spawn()
       be careful with vars! """
    digest_name = 'None'
    pid = os.getpid()
    while True:
        try:
            (exec_state, wu_arg1, wu_arg2) = q_work_unit.get(True)
            if exec_state == WorkerSignal.INIT:
                digest_constructor = wu_arg1
                digest_instance = digest_constructor()
                if digest_instance:
                    digest_name = digest_instance.name
                else:
                    digest_name = 'None'
                    q_debug.put((
                        logging.ERROR,
                        'worker_process({}) init -- pid={} Missing constructor'.format(
                            digest_name, pid)))
                q_debug.put((
                    logging.DEBUG,
                    'worker_process({}) init -- pid={}'.format(
                        digest_name, pid)))
                q_work_unit.task_done()
            elif exec_state == WorkerSignal.PROCESS:
                byte_block_len = wu_arg2
                if mmap_mode:
                    mm_block_name = wu_arg1
                    if byte_block_len > 0:
                        mm_block = mmap.mmap(
                            -1,
                            byte_block_len,
                            mm_block_name,
                        )
                        byte_block = mm_block.read(byte_block_len)
                        q_debug.put((
                            logging.DEBUG,
                            'worker_process() reading mmap -- pid={} l={} c={} d={} digest_instance={}'.format(
                                pid, byte_block_len, byte_block[0],
                                digest_instance.hexdigest(), mm_block_name)))
                    else:
                        byte_block = b''
                else:
                    byte_block = wu_arg1
                digest_instance.update(byte_block)
                q_debug.put((
                    logging.DEBUG,
                    'worker_process() process -- pid={} l={} d={}'.format(
                        pid, byte_block_len, digest_instance.hexdigest())))
                q_results.put('{} processed'.format(digest_name))
                q_work_unit.task_done()
            elif exec_state == WorkerSignal.RESULT:
                q_debug.put((
                    logging.DEBUG,
                    'worker_process({}) result -- pid={} digest={}'.format(
                        digest_name, pid, digest_instance.hexdigest())))
                q_results.put((digest_name, digest_instance.hexdigest()))
                q_work_unit.task_done()
            elif exec_state == WorkerSignal.QUIT:
                q_debug.put((
                    logging.DEBUG,
                    'worker_process({}) quit -- pid={}'.format(
                        digest_name, pid)))
                q_work_unit.task_done()
                break
            else:
                q_debug.put((
                    logging.WARNING,
                    'worker_process() bad state -- pid={}'.format(pid)))
                break  # bad state
        except KeyboardInterrupt:
            break



#!/usr/bin/env python3

"""

    Copyright (c) 2017 Martin F. Falatic

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

"""

import logging

PACKAGE_DATA = {
    'name': 'DirTreeDigest',
    'version': '0.6.5',
    'url': 'https://github.com/MartyMacGyver/DirTreeDigest',
    'author': 'Martin F. Falatic',
    'author_email': 'martin@falatic.com',
    'copyright': 'Copyright (c) 2017',
    'license': 'Apache License 2.0',
    'description': 'Directory Tree Digester',
    'long_description': """
A suite for digesting and analyzing filesystems:

dirtreedigest - generates cryptographic digests and collects stats across a directory tree

dirtreecmp - compare two dirtreedigest reports

This is a development release and is not considered final

For Windows, OS X and Linux
    """,
    'keywords': 'directory digest hashing integrity filesystem checksums',
    'classifiers': [
        'License :: OSI Approved :: Apache Software License',
        'Intended Audience :: Developers',
        'Topic :: Utilities',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        #'Development Status :: 3 - Alpha',
        'Development Status :: 4 - Beta',
        #'Development Status :: 5 - Production/Stable',
    ],
    'packages': [
        'dirtreedigest',
    ],
    'entry_points': {
        'console_scripts': [
            'dirtreedigest=dirtreedigest.main_digest:main',
            'dirtreecmp=dirtreedigest.main_compare:main',
        ],
    },
    'install_requires': [],
    'extras_require': {},
    'package_data': {},
    'data_files': [],
}

CONTROL_DATA = {
    'logfile_level': logging.INFO,
    'console_level': logging.INFO,
    'outfile_prefix': 'dtdigest',
    'outfile_ext': 'thd',
    'outfile_name': None,
    'altfile_name': None,
    'logfile_ext': 'log',
    'mmap_prefix': 'SHM_',
    'mmap_mode': True,
    'max_concurrent_jobs': 32,
    'max_buffers': 4,
    'max_block_size_mb': 16,
    'max_block_size': None,
    'ignore_path_case': False,
    'ignored_files': None,
    'ignored_dirs': None,
    'counts': {
        'files': 0,
        'dirs': 0,
        'ignored': 0,
        'errors': 0,
        'bytes_read': 0,
    },
    'altfile_digest': None,
    'buffer_blocks': None,
    'buffer_names': None,
    'buffer_sizes': None,
    'default_digests': None,
    'selected_digests': [],
    'p_worker_procs': None,
    'q_work_units': None,
    'q_results': None,
    'q_debug': None,
    'root_dir': None,
}

CONTROL_DATA['default_digests'] = [
    #'noop',
    #'crc32',
    #'adler32',
    'md5',
    'sha1',
    #'sha224',
    'sha256',
    #'sha384',
    #'sha512',
    #'blake2b',
    #'blake2s',
    #'sha3_224',
    'sha3_256',
    #'sha3_384',
    #'sha3_512',
    #'shake_128',
    #'shake_256',
    #'foo',
    #'sha1',
    #'noop1',
    #'noop2',
    #'noop3',
    #'noop4',
    #'noop5',
    #'noop6',
    #'noop7',
    #'noop8',
]

CONTROL_DATA['ignored_files'] = [
    'pagefile.sys',
    'hiberfil.sys',
]

CONTROL_DATA['ignored_dirs'] = [
    '$Recycle.Bin',
    'Recycled',
    'Recycler',
    'System Volume Information',
    'Temp',
    'msdownld.tmp',
]
